/***********************************************************

 Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

 SearchLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 SearchLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

 *************************************************************/

/*
 * CayleyManager.cpp
 *
 *  Created on: Sep 8, 2011
 *      Author: dstolee
 */

#include "SearchManager.hpp"
#include "CayleyManager.hpp"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#include "graph.h"
#include "cliquer.h"
#include "misc.h"
#include "reorder.h"
#include "set.h"

double seconds_per_clique_call = 1.0;
time_t last_print = 0;

int verbose = 0;
boolean killed_short;

int clique_count = 0;

boolean count_kill(set_t, graph_t *, clique_options *)
{
	clique_count++;

	if ( clique_count > 1 )
	{
		return FALSE;
	}

	return TRUE;
}

boolean short_kill(int level, int i, int n, int max, double putime, double realtime, clique_options *opts)
{
	if ( realtime <= seconds_per_clique_call )
	{
		return TRUE;
	}

	killed_short = TRUE;
	return FALSE;
}

int CayleyManager::largestClique()
{

	/* GRAPH METHODS */

	killed_short = FALSE;

	int max_size = clique_unweighted_max_weight(this->g, NULL);

	if ( killed_short == TRUE )
	{
		return -max_size;
	}

	return max_size;

	//	bzero(this->point_blocks, this->N * sizeof(int));
	//	this->num_available = this->N - 1;
	//
	//	for ( int j = 0; j < this->numGens; j++ )
	//	{
	//		int g = this->generators[j];
	//
	//		/* clear out spots! */
	//		this->num_available = this->num_available - 1;
	//		this->point_blocks[g] = this->N;
	//
	//		this->num_available = this->num_available - 1;
	//		this->point_blocks[this->N - g] = this->N;
	//	}
	//
	//	return this->largestClique(1, 1, 1);
}

int CayleyManager::largestClique(int min_v, int cur_size, int cur_max)
{
	if ( cur_size > cur_max )
	{
		cur_max = cur_size;
	}

	if ( min_v >= this->N )
	{

		return cur_max;
	}

	for ( int i = min_v; i < this->N; i++ )
	{
		if ( this->point_blocks[i] <= this->N - cur_size )
		{
			/* i is okay, don't block it! try i! */
			for ( int j = 0; j < this->numGens; j++ )
			{
				int g = this->generators[j];

				/* clear out spots! */
				if ( i + g < this->N && this->point_blocks[i + g] == 0 )
				{
					this->point_blocks[i + g] = this->N - cur_size;

				}

				if ( i + this->N - g < this->N && this->point_blocks[i + this->N - g] == 0 )
				{
					this->point_blocks[i + this->N - g] = this->N - cur_size;
				}
			}

			int num = this->largestClique(i + 1, cur_size + 1, cur_max);

			if ( num > cur_max )
			{
				cur_max = num;
			}

			/* i is okay, don't block it! try i! */
			for ( int j = 0; j < this->numGens; j++ )
			{
				int g = this->generators[j];

				/* reset spots! */
				if ( i + g < this->N && this->point_blocks[i + g] == this->N - cur_size )
				{
					this->point_blocks[i + g] = 0;

				}

				if ( i + this->N - g < this->N && this->point_blocks[i + this->N - g] == this->N - cur_size )
				{
					this->point_blocks[i + this->N - g] = 0;
				}
			}
		}
	}

	return cur_max;
}

int CayleyManager::numberCliques(int x, int y, int to_size)
{
	if ( x == y )
	{
		return -1;
	}

	GRAPH_ADD_EDGE(this->g, x, y);

	killed_short = FALSE;
	clique_count = 0;
	int num = clique_find_all(this->g, to_size, to_size, FALSE, NULL);

	GRAPH_DEL_EDGE(this->g, x, y);

	if ( killed_short == TRUE )
	{
		return -1 - num;
	}

	return num;
}

int CayleyManager::numberCliques(int x, int y, int to_size, int cur_size, int cur_num, int min_v)
{
	if ( cur_size == to_size )
	{
		return cur_num + 1;
	}

	if ( min_v >= this->N )
	{
		return cur_num;
	}

	for ( int i = min_v; i < this->N; i++ )
	{
		if ( i != x && i != y && this->point_blocks[i] == 0 )
		{
			this->clique[cur_size] = i;
			/* i is okay, don't block it! try i! */
			for ( int j = 0; j < this->numGens; j++ )
			{
				int g = this->generators[j];

				/* clear out spots! */
				if ( i + g < this->N && this->point_blocks[i + g] == 0 )
				{
					this->point_blocks[i + g] = cur_size;
				}

				if ( i + this->N - g < this->N && this->point_blocks[i + this->N - g] == 0 )
				{
					this->point_blocks[i + this->N - g] = cur_size;
				}
			}

			cur_num = this->numberCliques(x, y, to_size, cur_size + 1, cur_num, i + 1);

			if ( cur_num >= 2 )
			{
				return cur_num;
			}

			for ( int j = 0; j < this->numGens; j++ )
			{
				int g = this->generators[j];

				/* clear out spots! */
				if ( i + g < this->N && this->point_blocks[i + g] == cur_size )
				{
					this->point_blocks[i + g] = 0;
				}

				if ( i + this->N - g < this->N && this->point_blocks[i + this->N - g] == cur_size )
				{
					this->point_blocks[i + this->N - g] = 0;
				}
			}
		}
	}

	return cur_num;
}

void CayleyManager::init()
{
	this->numGens = this->maxGens;
	this->curGenIndex = 0;
	this->curMaxGenerator = 0;
	this->generators = 0;
	this->N = 0;
	this->R = 0;
	this->point_blocks = 0;

	if ( this->root != 0 )
	{
		delete this->root;
		this->root = new SearchNode(0);
	}

	this->generators = (int*) malloc(this->numGens * sizeof(int));
	this->generators[0] = 1;
	this->curMaxGenerator = 1;
	this->curGenIndex = 1;
}

int CayleyManager::clean()
{
	this->numGens = this->maxGens;
	this->curGenIndex = 0;

	if ( this->generators != 0 )
	{
		free(this->generators);
		this->generators = 0;
	}

	this->N = -1;
	this->R = -1;

	if ( this->point_blocks != 0 )
	{
		free(this->point_blocks);
		this->point_blocks = 0;
	}

	return 0;
}

/**
 * Constructor
 */
CayleyManager::CayleyManager(int maxGens, int maxN)
{
	this->maxGens = maxGens;
	this->maxN = maxN;

	this->init();
}

/**
 * Destructor
 */
CayleyManager::~CayleyManager()
{
	this->clean();
}

/**
 * pushNext -- deepen the search to the next child
 * 	of the current node.
 *
 * @return the label for the new node. -1 if none.
 */
LONG_T CayleyManager::pushNext()
{
	if ( this->numGens < 0 )
	{
		/* stage 1: find num gens */
		if ( this->root->curChild < 1 )
		{
			this->root->curChild = 1;
		}

		return this->pushTo(this->root->curChild + 1);
	}
	else if ( this->curGenIndex < this->numGens )
	{
		if ( this->stack.size() > 0 )
		{
			/* find a new generator! */
			SearchNode* parent = this->stack.back();

			if ( parent->curChild < 1 )
			{
				parent->curChild = this->curMaxGenerator;
			}

			return this->pushTo(parent->curChild + 1);
		}
		else
		{
			if ( this->root->curChild < 1 )
			{
				this->root->curChild = this->curMaxGenerator;
			}

			return this->pushTo(root->curChild + 1);
		}
	}
	else
	{
		/* N time! */
		SearchNode* parent = this->stack.back();

		if ( parent->curChild < 1 )
		{
			if ( 2 * this->curMaxGenerator < 16 )
			{
				parent->curChild = 10;
			}
			else
			{
				parent->curChild = 2 * this->curMaxGenerator - 1;
			}

		}

		return this->pushTo(parent->curChild + 1);
	}
}

/**
 * prune -- Perform a check to see if this nod
 * pushTo -- deepen the search to the specified child
 * 	of the current node.
 *
 * @param child the specified label for the new node
 * @return the label for the new node. -1 if none, or failed.
 */
LONG_T CayleyManager::pushTo(LONG_T child)
{
	if ( this->numGens < 0 )
	{
		/* stage 1: find num gens */
		if ( child > this->maxGens )
		{
			return -1;
		}

		this->numGens = child;

		this->generators = (int*) malloc(this->numGens * sizeof(int));
		this->generators[0] = 1;
		this->curMaxGenerator = 1;
		this->curGenIndex = 1;

		this->root->curChild = child;

		this->stack.push_back(new SearchNode(child));

		return child;
	}
	else if ( this->curGenIndex < this->numGens )
	{
		if ( child > this->maxN / 2 )
		{
			return -1;
		}

		this->generators[this->curGenIndex] = child;
		this->curGenIndex = this->curGenIndex + 1;
		this->curMaxGenerator = child;

		if ( this->stack.size() > 0 )
		{
			/* find a new generator! */
			SearchNode* parent = this->stack.back();
			parent->curChild = child;
		}
		else
		{
			this->root->curChild = child;
		}

		this->stack.push_back(new SearchNode(child));

		return child;
	}
	else
	{
		/* N time! */
		if ( this->N > 0 )
		{
			/* already pushed an N! */
			return -1;
		}

		if ( child > this->maxN )
		{
			return -1;
		}

		this->N = child;

		SearchNode* parent = this->stack.back();
		parent->curChild = child;

		this->stack.push_back(new SearchNode(child));

		return child;
	}
}

/**
 * pop -- remove the current node and move up the tree.
 *
 * @return the label of the node after the pop.
 * 		This return value is used for validation purposes
 * 		to check proper implementation of push*() and pop().
 */
LONG_T CayleyManager::pop()
{
	if ( this->numGens > 0 )
	{
		if ( this->curGenIndex <= 1 )
		{
			//			/* in numGens case */
			//			this->numGens = -1;
			//			this->N = -1;
			//			this->curGenIndex = 0;
			//			free(this->generators);
		}
		else if ( this->N > 0 )
		{
			/* in N case */
			this->N = -1;
		}
		else
		{
			this->N = -1;
			this->curGenIndex = this->curGenIndex - 1;
			this->curMaxGenerator = this->generators[this->curGenIndex - 1];
			this->generators[this->curGenIndex] = -1;
		}
	}

	if ( this->stack.size() > 0 )
	{
		SearchNode* parent = this->stack.back();
		this->stack.pop_back();

		LONG_T label = parent->label;

		delete parent;

		return label;
	}
	return this->root->curChild;
}

/**
 * prune -- Perform a check to see if this node should be pruned.
 *
 * @return 0 if no prune should occur, 1 if prune should occur.
 */
int CayleyManager::prune()
{
	return 0;
}

/**
 * isSolution -- Perform a check to see if a solution exists
 * 		at this point.
 *
 * @return 0 if no solution is found, 1 if a solution is found.
 */
int CayleyManager::isSolution()
{
	if ( this->numGens > 0 && this->curGenIndex >= this->numGens && this->N > 0 )
	{
		if ( (this->deepeningMode & DEEPEN_GENERATE) != 0 )
		{
			(this->searchDepth)++;
			this->writeJob(stdout);
			(this->searchDepth)--;
			return 0;
		}

		double old_kill_time = seconds_per_clique_call;
		if ( 0 && this->jobDepth >= this->searchDepth )
		{
			/* we need to use a different kill time! */
			int cur_time = time(NULL);
			seconds_per_clique_call = (this->end_time - cur_time) / (1 + this->numGens);
			printf("-- using clique time %d (%d)\n", seconds_per_clique_call, 1 + (this->numGens));
		}

		/* TOUGH PART! */
		if ( verbose > 0 )
		{
			printf("-- N %d...", this->N);
			for ( int i = 0; i < this->numGens; i++ )
			{
				printf("g%d=%d...", i, this->generators[i]);
			}
			printf("\n");
			fflush(stdout);
		}
		this->g = graph_new(this->N);

		clique_default_options->output = NULL;
		clique_default_options->time_function = &short_kill;
		clique_default_options->user_function = &count_kill;

		for ( int i = 0; i < this->N; i++ )
		{
			int cgen = 0;
			for ( int j = 1; j <= (this->N + 1) / 2; j++ )
			{
				if ( cgen < this->numGens && j == this->generators[cgen] )
				{
					cgen++;
				}
				else
				{
					GRAPH_ADD_EDGE(this->g, i, (i+j) % this->N);
				}
			}
		}

		int max_clique = this->largestClique();

		if ( max_clique < 2 )
		{
			/* killed short! */
			if ( verbose > 0 )
			{
				printf("halted while computing clique number.\n");
				fflush(stdout);
			}
			(this->searchDepth)++;
			this->writeJob(stdout);
			(this->searchDepth)--;

			seconds_per_clique_call = old_kill_time;
			return 0;
		}

		this->R = max_clique + 1;
		if ( verbose > 0 )
		{
			printf("--N %d R %d\n", this->N, this->R);
			fflush(stdout);
		}

		bool possible = true;

		for ( int i = 0; possible && i < this->numGens; i++ )
		{
			clique_count = 0;
			int nc = this->numberCliques(0, this->generators[i], this->R);

			if ( nc < 0 )
			{
				/* killed short! */
				if ( verbose > 0 )
				{
					printf("halted while counting cliques in G + { 0, %d } (the generator of index %d).\n",
							this->generators[i], i);
					fflush(stdout);
				}

				(this->searchDepth)++;
				this->writeJob(stdout);
				(this->searchDepth)--;

				seconds_per_clique_call = old_kill_time;
				return 0;
			}

			if ( verbose > 0 )
			{
				printf("Checked generator %d...\n", this->generators[i]);
				fflush(stdout);
			}

			if ( nc != 1 )
			{
				/* NO! Either no completion or more than one. */
				if ( verbose > 0 )
				{
					printf("There are %d cliques in G + { 0, %d } (the generator of index %d).\n", nc,
							this->generators[i], i);
					fflush(stdout);
				}

				//		printf("(%d cliques is bad!)...", nc);
				possible = false;
			}
		}

		graph_free(this->g);
		this->g = 0;

		//printf("done\n");
		seconds_per_clique_call = old_kill_time;

		if ( possible )
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}

	return 0;
}
/**
 * writeSolution -- create a buffer that contains a
 * 	description of the solution.
 *
 * Solution strings start with S.
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* CayleyManager::writeSolution()
{
	printf("N = %5d\tR = %2d\tG = %d\tGenerators = { ", this->N, this->R, this->numGens);
	for ( int i = 0; i < this->numGens; i++ )
	{
		printf("%3d ", this->generators[i]);
	}
	printf("}\n");
	return 0;
}

/**
 * writeStatistics -- create a buffer that contains a
 * 	description of the solution.
 *
 * Statistics take the following format in each line:
 *
 * T [TYPE] [ID] [VALUE]
 *
 * @return a string of data allocated with malloc().
 * 	It will be deleted using free().
 */
char* CayleyManager::writeStatistics()
{
	return super::writeBaseStats();
}

void CayleyManager::reset()
{
	this->clean();
	this->init();
}
