/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * CliquerData.hpp
 *
 *  Created on: Sep 11, 2011
 *      Author: derrickstolee
 */

#ifndef CLIQUERDATA_HPP_
#define CLIQUERDATA_HPP_

#include <stack>
#include "Set.hpp"
#include "cliquer.h"
#include "graph.h"
#include "GraphData.hpp"
#include "nausparse.h"

class CliquerData: public GraphData
{
protected:
	typedef GraphData super;

	/**
	 * clique_graph
	 */
	graph_t* cliquer_graph;
	void initCliquerGraph();
	void cleanCliquerGraph();
	void addEdgeToCliquerGraph(int i, int j);
	std::stack<int> cliquer_graph_stack;
	std::stack<int> cliquer_graph_sizes;
	void snapshotCliquerGraph();
	void rollbackCliquerGraph();

	virtual void init();
	virtual void clean();

public:
	/**
	 * Constructor
	 */
	CliquerData(int r, int n);

	/**
	 * Destructor
	 */
	virtual ~CliquerData();

	/**
	 * snapshot()
	 *
	 * Freeze the current state in a stack.
	 */
	virtual void snapshot();

	/**
	 * rollback()
	 *
	 * Pop the stack and revert the structure to the previous state.
	 */
	virtual void rollback();

	/**
	 * reset()
	 */
	virtual void reset();

	/**
	 * augmentEdge()
	 *
	 * Augment the graph using the following edge change.
	 *
	 * return FALSE if the constraints are violated.
	 */
	bool augmentEdge(int i, int j, char type);

	/**
	 * getCliquerGraph()
	 */
	graph_t* getCliquerGraph();

	/**
	 * isConsistent
	 */
	virtual bool isConsistent();

	/**
	 * isUniquelyKrSaturated
	 */
	virtual bool isUniquelyKrSaturated();

};

#endif /* CLIQUERDATA_HPP_ */
