/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * SBRConnected.hpp
 *
 *  Created on: Aug 24, 2011
 *      Author: derrickstolee
 */

#ifndef SBRCONNECTED_HPP_
#define SBRCONNECTED_HPP_

#include "GraphData.hpp"
#include "SaturationNode.hpp"
#include "SaturationBranchingRule.hpp"

typedef enum
{
	BIGGEST, MAX_DEGREE
} SubBranchingRule;

class SBRConnected: public SaturationBranchingRule
{
protected:
	SubBranchingRule sub;

public:
	/**
	 * Constructor
	 *
	 * Takes an enum for the "sub" branching rule: how to select within the "connected" classes.
	 */
	SBRConnected(SubBranchingRule rule);

	virtual ~SBRConnected();

	/**
	 * selectOrbitIndex
	 *
	 * Perform the implemented branching rule to select an orbit
	 * 	index from the given list of orbits.
	 * The input orbits should be orbits corresponding to pair-orbits
	 * 	of "unknown" edges in the current partial graph.
	 *
	 * @param data the GraphData object containing all of the graph data.
	 * @param node the SaturationNode object containing all of the orbit data.
	 *
	 */
	virtual int selectOrbitIndex(GraphData* data, SaturationNode* node);
};

#endif /* SBRCONNECTED_HPP_ */
