/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * SBRLargestOrbit.cpp
 *
 *  Created on: Aug 24, 2011
 *      Author: dstolee
 */

#include "SaturationData.hpp"
#include "SaturationNode.hpp"
#include "SaturationBranchingRule.hpp"
#include "SBRLargestOrbit.hpp"

/**
 * Constructor... not really important.
 */
SBRLargestOrbit::SBRLargestOrbit()
{

}

/**
 * Destructor!
 */
SBRLargestOrbit::~SBRLargestOrbit()
{

}

/**
 * selectOrbitIndex
 *
 * Perform the implemented branching rule to select an orbit
 * 	index from the given list of orbits.
 * The input orbits should be orbits corresponding to pair-orbits
 * 	of "unknown" edges in the current partial graph.
 *
 * @param data the GraphData object containing all of the graph data.
 * @param node the SaturationNode object containing all of the orbit data.
 *
 */
int SBRLargestOrbit::selectOrbitIndex(GraphData* data, SaturationNode* node)
{
	int best_index = 0;
	int best_size = node->orbitSizes[0];

	for ( int i = 1; i < node->numOrbits; i++ )
	{
		if ( node->orbitSizes[i] > best_size )
		{
			best_index = i;
			best_size = node->orbitSizes[i];
		}
	}

	return best_index;
}
