/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * cayley.cpp
 *
 *  Created on: Sep 8, 2011
 *      Author: dstolee
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "SearchManager.hpp"
#include "CayleyManager.hpp"

extern double seconds_per_clique_call;
extern int verbose;

int main(int argc, char** argv)
{
	int maxGens = -1;
	int maxN = -1;

	for ( int i = 1; i < argc; i++ )
	{
		if ( argv[i][0] == '-' && argv[i][1] != 0 && argv[i][2] == 0  && i < argc - 1)
		{
			if ( argv[i][1] == 'G' )
			{
				maxGens = atoi(argv[i + 1]);
			}
			else if ( argv[i][1] == 'N' )
			{
				maxN = atoi(argv[i + 1]);
			}
			else if ( argv[i][1] == 't' )
			{
				seconds_per_clique_call = (double)atof(argv[i+1]);
			}
		}

		if ( strcmp(argv[i],"--verbose") == 0 )
		{
			verbose = 1;
		}
	}

	if ( maxGens < 0 || maxN < 0 )
	{
		printf("Usage: cayley.exe -G # -N # [TreeSearch args]\n");
		exit(1);
	}

	CayleyManager* manager = new CayleyManager(maxGens, maxN);
	manager->importArguments(argc, argv);

	while ( manager->readJob(stdin) >= 0 )
	{
		manager->doSearch();
		manager->reset();
	}

	delete manager;

	return 0;

}
