/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * graphcompactor.c
 *
 *  Created on: Apr 18, 2011
 *      Author: stolee
 */

#include <stdlib.h>
#include "nausparse.h"
#include "gtools.h"
#include "graphcompactor.h"

/**
 * compactsg
 *
 * Given a sparsegraph which may have gaps in its representation (the e array),
 * 	compact it to have no gaps.
 */
sparsegraph* compactsg( sparsegraph* g )
{
	sparsegraph* cg = (sparsegraph*) malloc(sizeof(sparsegraph));

	SG_INIT(*cg);

	cg->nv = g->nv;
	cg->vlen = g->nv;
	cg->dlen = g->nv;

	cg->v = (int*) malloc(cg->vlen * sizeof(int));
	cg->d = (int*) malloc(cg->dlen * sizeof(int));

	int i, j;
	int vindex = 0;
	for ( i = 0; i < g->nv; i++ )
	{
		cg->v[i] = vindex;
		cg->d[i] = g->d[i];
		vindex += g->d[i];
	}

	cg->nde = vindex;
	cg->elen = vindex;
	cg->e = (int*) malloc(cg->elen * sizeof(int));

	for ( i = 0; i < g->nv; i++ )
	{
		vindex = cg->v[i];
		int gvindex = g->v[i];
		int d = g->d[i];

		for ( j = 0; j < d; j++ )
		{
			cg->e[vindex + j] = g->e[gvindex + j];
		}
	}

	return cg;
}
