/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * paircycles.cpp
 *
 *  Created on: Oct 15, 2011
 *      Author: dstolee
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "graph.h"
#include "cliquer.h"
#include "misc.h"
#include "reorder.h"
#include "set.h"
#include "translation.hpp"

time_t last_print = 0;

boolean killed_short;

int clique_count = 0;

double seconds_per_clique_call = 100;

boolean count_kill(set_t, graph_t *, clique_options *)
{
	clique_count++;

	if ( clique_count > 1 )
	{
		killed_short = TRUE;
		return FALSE;
	}

	return TRUE;
}

boolean short_kill(int level, int i, int n, int max, double putime, double realtime, clique_options *opts)
{
	if ( realtime <= seconds_per_clique_call )
	{
		return TRUE;
	}

	killed_short = TRUE;
	printf("--killed short for time\n");
	return FALSE;
}

int main(int argc, char** argv)
{
	int N = 5;

	graph_t* g = graph_new(100);

	while ( scanf("%d", &N) != EOF )
	{
		int num_gens;
		scanf("%d", &num_gens);

		int* S = (int*) malloc(num_gens * sizeof(int));
		int* T = (int*) malloc(N * sizeof(int));

		for ( int j = 0; j < num_gens; j++ )
		{
			scanf("%d", &(S[j]));
		}

		killed_short = FALSE;

		seconds_per_clique_call = 1000;

		for ( int num_elts = 1; num_elts <= N - 1; num_elts++ )
		{
			for ( int set_index = 0; set_index < nChooseK(N, num_elts); set_index++ )
			{
				for ( int rot = 0; rot < N; rot++ )
				{

					indexToSet(num_elts, set_index, T);

					clique_default_options->output = NULL;
					clique_default_options->time_function = &short_kill;
					clique_default_options->user_function = &count_kill;

					GRAPH_ADD_EDGE(g, 0, 1);

					for ( int i = 0; i < N; i++ )
					{
						GRAPH_ADD_EDGE(g, 0, 2 + i);
						GRAPH_ADD_EDGE(g, 1, 2 + N + i);

						int cur_index = 0;
						for ( int j = 0; j < num_gens; j++ )
						{
							GRAPH_ADD_EDGE(g, 2 + i, 2 + ((i + S[j]) % N));
							GRAPH_ADD_EDGE(g, 2 + N + i, 2 + N + ((i + S[j]) % N));
						}

						for ( int j = 0; j < num_elts; j++ )
						{
							GRAPH_ADD_EDGE(g, 2 + i, 2 + N + ((rot * i + T[j]) % N));
						}
					}

					int max_size = clique_unweighted_max_weight(g, NULL) + 1;

					bool is_unique = true;

					int cur_index = 0;
					for ( int j = 0; is_unique && j < N; j++ )
					{
						if ( cur_index < num_elts && T[cur_index] == j )
						{
							cur_index++;
						}
						else
						{
							GRAPH_ADD_EDGE(g, 2, 2 + N + j);

							killed_short = FALSE;
							clique_count = 0;
							int num = clique_find_all(g, max_size, max_size, FALSE, NULL);

							GRAPH_DEL_EDGE(g, 2, 2 + N + j);

							if ( num != 1 )
							{
								is_unique = false;
//							printf("-- failed on 2,%d\t (j = %d)\n", 2 + N + j, j);
							}

							if ( killed_short == TRUE )
							{
								is_unique = false;
								killed_short = FALSE;
							}
						}
					}

					if ( is_unique )
					{
						GRAPH_ADD_EDGE(g, 0, 2 + N);

						killed_short = FALSE;
						clique_count = 0;
						int num = clique_find_all(g, max_size, max_size, FALSE, NULL);

						GRAPH_DEL_EDGE(g, 0, 2 + N);

						if ( num != 1 )
						{
							is_unique = false;
//						printf("-- failed on 0,%d\n", 2 + N);
						}

						if ( killed_short == TRUE )
						{
							is_unique = false;
							killed_short = FALSE;
						}
					}

					if ( is_unique )
					{
						GRAPH_ADD_EDGE(g, 1, 2);

						killed_short = FALSE;
						clique_count = 0;
						int num = clique_find_all(g, max_size, max_size, FALSE, NULL);

						GRAPH_DEL_EDGE(g, 1, 2);

						if ( num != 1 )
						{
							is_unique = false;
//						printf("-- failed on 1,2\n");
						}

						if ( killed_short == TRUE )
						{
							is_unique = false;
							killed_short = FALSE;
						}
					}

					if ( is_unique )
					{
						printf("-- YAY. %d %d %d\t", N, num_elts, rot);

						for ( int j = 0; j < num_elts; j++ )
						{
							printf("%d ", T[j]);
						}
						printf("\n");
					}

					for ( int i = 0; i < 2 * N + 2; i++ )
					{
						for ( int j = i + 1; j < 2 * N + 2; j++ )
						{
							GRAPH_DEL_EDGE(g, i, j);
						}
					}

				} /* end for all rots */
			} /* end for all sets T */
		} /* end for all set sizes */

		free(S);
		free(T);
		printf("-------------------------------------------------------- done with N = %d\n", N);
	}
}

