/***********************************************************

Copyright Derrick Stolee 2011.

 This file is part of SearchLib.

    SearchLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SearchLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SearchLib.  If not, see <http://www.gnu.org/licenses/>.

*************************************************************/

/*
 * saturation.cpp
 *
 *  Created on: Apr 7, 2011
 *      Author: stolee
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "SaturationData.hpp"
#include "SearchManager.hpp"
#include "SaturationBranchingRule.hpp"
#include "SBRConnected.hpp"
#include "SBRLargestOrbit.hpp"
#include "OrbitalBranchingManager.hpp"

#define _MODE_CANONICAL_ -1
#define _MODE_ORBITAL_ 1

#define _INIT_NONE_ 0
#define _INIT_GR6N20A_ 1

int main(int argc, char** argv)
{
	int init_mode = 0;
	int r = -1;
	int N = -1;
	int mode = _MODE_ORBITAL_;
	SaturationBranchingRule* sbr = 0;
	bool use_cliquer = false;

	int i;
	for ( i = 1; i < argc - 1; i++ )
	{
		if ( argv[i][0] == '-' && argv[i][1] != 0 && argv[i][2] == 0 )
		{
			if ( argv[i][1] == 'r' )
			{
				r = atoi(argv[i + 1]);
			}
			else if ( argv[i][1] == 'N' )
			{
				N = atoi(argv[i + 1]);
			}
		}
		else if ( strcmp(argv[i], "--mode") == 0 )
		{
			if ( strcmp(argv[i + 1], "canon") == 0 )
			{
				mode = _MODE_CANONICAL_;
			}
			else if ( strcmp(argv[i + 1], "orbital") == 0 )
			{
				mode = _MODE_ORBITAL_;
			}
		}
		else if ( strcmp(argv[i], "--rule") == 0 )
		{
			if ( strcmp(argv[i + 1], "biggest") == 0 )
			{
				sbr = new SBRLargestOrbit();
			}
			else if ( strcmp(argv[i + 1], "connected_big") == 0 )
			{
				sbr = new SBRConnected(BIGGEST);
			}
			else if ( strcmp(argv[i + 1], "connected_deg") == 0 )
			{
				sbr = new SBRConnected(MAX_DEGREE);
			}
		}
		else if ( strcmp(argv[i], "--init") == 0 )
		{
			if ( strcmp(argv[i + 1], "gr6n20a") == 0 )
			{
				init_mode = _INIT_GR6N20A_;
			}
		}
		else if ( strcmp(argv[i], "--cliquer") == 0 )
		{
			use_cliquer = true;
		}
	}

	if ( strcmp(argv[argc-1], "--cliquer") == 0 )
	{
		use_cliquer = true;
	}

	if ( r < 0 || N < 0 )
	{
		printf("Usage: saturation.exe -r # -N # [TreeSearch args]\n");
		exit(1);
	}

	SearchManager* manager = 0;

	if ( sbr == 0 )
	{
		sbr = new SBRLargestOrbit();
	}

	if ( mode == _MODE_CANONICAL_ )
	{
		//		manager = new SaturationAugmenter(r, N, mode);
	}
	else if ( mode == _MODE_ORBITAL_ )
	{
		manager = new OrbitalBranchingManager(r, N, sbr, use_cliquer);
	}

	manager->importArguments(argc, argv);

	while ( manager->readJob(stdin) >= 0 )
	{
		switch ( init_mode )
		{
		case _INIT_GR6N20A_:
			((OrbitalBranchingManager*) manager)->initToGR6N20A();
			break;

		}

		manager->doSearch();

		if ( mode == _MODE_ORBITAL_ )
		{
			((OrbitalBranchingManager*) manager)->reset();
		}
	}

	delete manager;
	delete sbr;

	return 0;
}
